@@ -2,26 +2,24 @@
 # this repository contains the full copyright notices and license terms.
 import datetime
 from decimal import Decimal
-from itertools import groupby
 from functools import partial
+from itertools import groupby
 
 from trytond.i18n import gettext
-from trytond.model import ModelView, ModelSQL, fields
+from trytond.model import Index, ModelSQL, ModelView, fields
 from trytond.model.exceptions import AccessError
-from trytond.wizard import Wizard, StateView, StateTransition, Button
-from trytond.pyson import If, Eval, Bool
-from trytond.tools import sortable_values, firstline
-from trytond.transaction import Transaction
-from trytond.pool import Pool
-
 from trytond.modules.company.model import (
-    employee_field, set_employee, reset_employee)
+    employee_field, reset_employee, set_employee)
 from trytond.modules.product import round_price
+from trytond.pool import Pool
+from trytond.pyson import Bool, Eval, If
+from trytond.tools import firstline, sortable_values
+from trytond.transaction import Transaction
+from trytond.wizard import Button, StateTransition, StateView, Wizard
 
 STATES = {
     'readonly': Eval('state') != 'draft',
     }
-DEPENDS = ['state']
 
 
 class PurchaseRequest(ModelSQL, ModelView):
@@ -29,24 +27,29 @@ class PurchaseRequest(ModelSQL, ModelView):
     __name__ = 'purchase.request'
 
     product = fields.Many2One(
-        'product.product', "Product", select=True, readonly=True,
-        domain=[('purchasable', '=', True)],
+        'product.product', "Product", readonly=True,
+        domain=[
+            If((Eval('state') == 'draft')
+                & ~(Eval('quantity', 0) < 0),
+                ('purchasable', '=', True),
+                ()),
+            ],
         context={
             'company': Eval('company', -1),
             },
-        depends=['company'])
+        depends={'company'})
     description = fields.Text('Description', readonly=True)
     summary = fields.Function(fields.Char('Summary'), 'on_change_with_summary')
     party = fields.Many2One(
-        'party.party', "Party", select=True, states=STATES,
+        'party.party', "Party", states=STATES,
         context={
             'company': Eval('company', -1),
             },
-        depends=DEPENDS + ['company'])
-    quantity = fields.Float('Quantity', required=True, states=STATES,
-        digits=(16, Eval('uom_digits', 2)), depends=DEPENDS + ['uom_digits'])
-    uom = fields.Many2One('product.uom', 'UOM', select=True,
-        ondelete='RESTRICT',
+        depends={'company'})
+    quantity = fields.Float(
+        "Quantity", required=True, states=STATES, digits='unit')
+    unit = fields.Many2One(
+        'product.uom', "Unit", ondelete='RESTRICT',
         domain=[
             If(Bool(Eval('product_uom_category')),
                 ('category', '=', Eval('product_uom_category')),
@@ -55,31 +58,30 @@ class PurchaseRequest(ModelSQL, ModelView):
         states={
             'required': Bool(Eval('product')),
             'readonly': STATES['readonly'],
-            },
-        depends=['product', 'product_uom_category'] + DEPENDS)
-    uom_digits = fields.Function(fields.Integer('UOM Digits'),
-        'on_change_with_uom_digits')
+            })
     product_uom_category = fields.Function(
-        fields.Many2One('product.uom.category', "Product Uom Category"),
+        fields.Many2One(
+            'product.uom.category', "Product UoM Category",
+            help="The category of Unit of Measure for the product."),
         'on_change_with_product_uom_category')
     computed_quantity = fields.Float('Computed Quantity', readonly=True)
-    computed_uom = fields.Many2One('product.uom', 'Computed UOM',
-        readonly=True)
+    computed_unit = fields.Many2One(
+        'product.uom', "Computed Unit", readonly=True)
     purchase_date = fields.Date('Best Purchase Date', readonly=True)
     supply_date = fields.Date('Expected Supply Date', readonly=True)
-    default_uom = fields.Function(fields.Many2One(
-            'product.uom', "Default UOM"), 'on_change_with_default_uom')
-    default_uom_digits = fields.Function(fields.Integer('Default UOM Digits'),
-        'on_change_with_default_uom_digits')
-    stock_level = fields.Float('Stock at Supply Date', readonly=True,
-        digits=(16, Eval('default_uom_digits', 2)),
-        depends=['default_uom_digits'])
+    default_uom = fields.Function(
+        fields.Many2One(
+            'product.uom', "Default UoM",
+            help="The default Unit of Measure."),
+        'on_change_with_default_uom')
+    stock_level = fields.Float(
+        "Stock at Supply Date", readonly=True, digits='default_uom')
     warehouse = fields.Many2One(
         'stock.location', "Warehouse",
         states={
             'required': Eval('warehouse_required', False),
             },
-        domain=[('type', '=', 'warehouse')], depends=['warehouse_required'],
+        domain=[('type', '=', 'warehouse')],
         readonly=True)
     warehouse_required = fields.Function(fields.Boolean('Warehouse Required'),
         'get_warehouse_required')
@@ -95,16 +97,22 @@ class PurchaseRequest(ModelSQL, ModelView):
     purchased_by = employee_field(
         "Purchased By", states=['purchased', 'done', 'cancelled', 'exception'])
     state = fields.Selection([
+            ('draft', "Draft"),
             ('purchased', "Purchased"),
             ('done', "Done"),
-            ('draft', "Draft"),
             ('cancelled', "Cancelled"),
             ('exception', "Exception"),
-            ], "State", required=True, readonly=True, select=True)
+            ], "State", required=True, readonly=True, sort=False)
 
     @classmethod
     def __setup__(cls):
         super(PurchaseRequest, cls).__setup__()
+        t = cls.__table__()
+        cls._sql_indexes.add(
+            Index(
+                t,
+                (t.state, Index.Equality()),
+                where=t.state.in_(['draft', 'purchased', 'exception'])))
         cls._order[0] = ('id', 'DESC')
         cls._buttons.update({
                 'handle_purchase_cancellation_exception': {
@@ -115,63 +123,22 @@ class PurchaseRequest(ModelSQL, ModelView):
 
     @classmethod
     def __register__(cls, module_name):
-        pool = Pool()
-        ModelData = pool.get('ir.model.data')
-        Purchase = pool.get('purchase.purchase')
-        PurchaseLine = pool.get('purchase.line')
-        model_data = ModelData.__table__()
-        purchase = Purchase.__table__()
-        purchase_line = PurchaseLine.__table__()
         request = cls.__table__()
+        table_h = cls.__table_handler__(module_name)
 
-        tablehandler = cls.__table_handler__(module_name)
-        state_exist = tablehandler.column_exist('state')
+        # Migration from 6.8: rename uom to unit
+        if (table_h.column_exist('uom')
+                and not table_h.column_exist('unit')):
+            table_h.column_rename('uom', 'unit')
+
+        # Migration from 6.8: rename uom to unit
+        if (table_h.column_exist('computed_uom')
+                and not table_h.column_exist('computed_unit')):
+            table_h.column_rename('computed_uom', 'computed_unit')
 
         super(PurchaseRequest, cls).__register__(module_name)
 
-        # Migration from 3.6: removing the constraint on the quantity
-        tablehandler = cls.__table_handler__(module_name)
-        tablehandler.drop_constraint('check_purchase_request_quantity')
-
-        # Migration from 3.8: renaming module of Purchase Request group entry
         cursor = Transaction().connection.cursor()
-        cursor.execute(*model_data.update(
-                columns=[model_data.module],
-                values=['purchase_request'],
-                where=((model_data.fs_id == 'group_purchase_request')
-                    & (model_data.module == 'stock_supply'))))
-
-        # Migration from 4.0: remove required on product and uom
-        tablehandler.not_null_action('product', action='remove')
-        tablehandler.not_null_action('uom', action='remove')
-
-        # Migration from 4.2: add state
-        if not state_exist:
-            cursor = Transaction().connection.cursor()
-            update = Transaction().connection.cursor()
-            query = request.join(purchase_line, type_='INNER',
-                condition=request.purchase_line == purchase_line.id
-                ).join(purchase, type_='INNER',
-                    condition=purchase_line.purchase == purchase.id
-                    ).select(
-                        request.id, purchase.state, request.exception_ignored)
-            cursor.execute(*query)
-            for request_id, purchase_state, exception_ignored in cursor:
-                if purchase_state == 'cancel' and not exception_ignored:
-                    state = 'exception'
-                elif purchase_state == 'cancel':
-                    state = 'cancel'
-                elif purchase_state == 'done':
-                    state = 'done'
-                else:
-                    state = 'purchased'
-                update.execute(*request.update(
-                        [request.state],
-                        [state],
-                        where=request.id == request_id))
-
-        # Migration from 4.4: remove required on origin
-        tablehandler.not_null_action('origin', action='remove')
 
         # Migration from 5.6: rename state cancel to cancelled
         cursor.execute(*request.update(
@@ -183,9 +150,9 @@ class PurchaseRequest(ModelSQL, ModelView):
         Lang = pool.get('ir.lang')
         if self.product:
             lang = Lang.get()
-            rec_name = (lang.format(
-                    '%.*f', (self.uom.digits, self.quantity))
-                + '%s %s' % (self.uom.symbol, self.product.name))
+            rec_name = (lang.format_number_symbol(
+                    self.quantity, self.unit, digits=self.unit.digits)
+                + ' %s' % self.product.name)
         elif self.description:
             rec_name = self.description.splitlines()[0]
         else:
@@ -233,7 +200,10 @@ class PurchaseRequest(ModelSQL, ModelView):
 
     @property
     def currency(self):
-        return self.company.currency
+        currency = self.company.currency
+        if self.party and self.party.supplier_currency:
+            currency = self.party.supplier_currency
+        return currency
 
     @classmethod
     def default_state(cls):
@@ -273,27 +243,13 @@ class PurchaseRequest(ModelSQL, ModelView):
     def get_warehouse_required(self, name):
         return self.product and self.product.type in ('goods', 'assets')
 
-    @fields.depends('uom')
-    def on_change_with_uom_digits(self, name=None):
-        if self.uom:
-            return self.uom.digits
-        return 2
-
     @fields.depends('product')
     def on_change_with_product_uom_category(self, name=None):
-        if self.product:
-            return self.product.default_uom_category.id
+        return self.product.default_uom_category if self.product else None
 
     @fields.depends('product')
     def on_change_with_default_uom(self, name=None):
-        if self.product:
-            return self.product.default_uom.id
-
-    @fields.depends('product')
-    def on_change_with_default_uom_digits(self, name=None):
-        if self.product:
-            return self.product.default_uom.digits
-        return 2
+        return self.product.default_uom if self.product else None
 
     @fields.depends('description')
     def on_change_with_summary(self, name=None):
@@ -320,11 +276,10 @@ class PurchaseRequest(ModelSQL, ModelView):
 
     @classmethod
     def create(cls, vlist):
-        for vals in vlist:
-            for field_name in ('quantity', 'company'):
-                if vals.get(field_name) is None:
-                    raise AccessError(
-                        gettext('purchase_request.msg_request_no_create'))
+        transaction = Transaction()
+        if transaction.user and transaction.check_access:
+            raise AccessError(
+                gettext('purchase_request.msg_request_no_create'))
         return super(PurchaseRequest, cls).create(vlist)
 
     @classmethod
@@ -350,7 +305,8 @@ class PurchaseRequest(ModelSQL, ModelView):
         "Return the best product supplier to request product at date"
         pool = Pool()
         Date = pool.get('ir.date')
-        today = Date.today()
+        with Transaction().set_context(context=product._context):
+            today = Date.today()
         for product_supplier in product.product_suppliers_used(**pattern):
             if date is None:
                 return product_supplier
@@ -374,7 +330,8 @@ class PurchaseRequest(ModelSQL, ModelView):
             purchase_date = product_supplier.compute_purchase_date(date)
         else:
             supplier = None
-            purchase_date = Date.today()
+            with Transaction().set_context(context=product._context):
+                purchase_date = Date.today()
         return supplier, purchase_date
 
     @classmethod
@@ -438,7 +395,7 @@ class CreatePurchase(Wizard):
         return (
             ('product', request.product),
             ('description', request.description),
-            ('unit', request.uom),
+            ('unit', request.unit),
             )
 
     def transition_start(self):
@@ -477,8 +434,6 @@ class CreatePurchase(Wizard):
         if any(reqs):
             return 'ask_party'
 
-        today = Date.today()
-
         requests = [r for r in requests if not r.purchase_line]
 
         def _sort_keyfunc(requests, request):
@@ -494,6 +449,9 @@ class CreatePurchase(Wizard):
         lines = []
         for key, grouped_requests in groupby(requests, key=keyfunc):
             grouped_requests = list(grouped_requests)
+            key = dict(key)
+            with Transaction().set_context(company=key['company']):
+                today = Date.today()
             try:
                 purchase_date = min(r.purchase_date
                     for r in grouped_requests
@@ -503,7 +461,7 @@ class CreatePurchase(Wizard):
             if purchase_date < today:
                 purchase_date = today
             purchase = Purchase(purchase_date=purchase_date)
-            for f, v in key:
+            for f, v in key.items():
                 setattr(purchase, f, v)
             purchases.append(purchase)
             for line_key, line_requests in groupby(
@@ -525,18 +483,34 @@ class CreatePurchase(Wizard):
         Line = pool.get('purchase.line')
 
         line = Line()
-        line.unit_price = round_price(Decimal(0))
         for f, v in key:
             setattr(line, f, v)
-        line.quantity = sum(r.quantity for r in requests)
         line.purchase = purchase
         line.on_change_product()
+        line.quantity = cls.compute_quantity(requests, line, purchase)
+        if line.unit:
+            line.quantity = line.unit.ceil(line.quantity)
         # Set again in case on_change's changed them
         for f, v in key:
             setattr(line, f, v)
         line.on_change_quantity()
+        if line.unit_price is None:
+            line.unit_price = round_price(Decimal(0))
         return line
 
+    @classmethod
+    def compute_quantity(cls, requests, line, purchase):
+        pool = Pool()
+        Uom = pool.get('product.uom')
+        unit = line.unit
+        compute_qty = Uom.compute_qty
+        return sum(
+            compute_qty(r.unit, r.quantity, unit, round=False)
+            for r in requests)
+
+    def end(self):
+        return 'reload'
+
 
 class HandlePurchaseCancellationException(Wizard):
     'Handle Purchase Cancellation Exception'
